import os
import sys
import socket
from threading import Thread, Lock
import logging
import resource
import papa
from papa.utils import Error, cast_bytes, cast_string, recv_with_retry, send_with_retry
from papa.server import papa_socket, values, proc
import atexit
try:
    # noinspection PyPackageRequirements
    from setproctitle import setproctitle
except ImportError:
    setproctitle = None

__author__ = 'Scott Maxwell'

try:
    from argparse import ArgumentParser
except ImportError:
    from optparse import OptionParser

    class ArgumentParser(OptionParser):
        def add_argument(self, *args, **kwargs):
            return self.add_option(*args, **kwargs)

        # noinspection PyShadowingNames
        def parse_args(self, args=None, values=None):
            return OptionParser.parse_args(self, args, values)[0]

log = logging.getLogger('papa.server')


class CloseSocket(Exception):
    def __init__(self, final_message=None):
        self.final_message = final_message
        super(CloseSocket, self).__init__(self)


# noinspection PyUnusedLocal
def quit_command(sock, args, instance):
    """Close the client socket"""
    raise CloseSocket('ok\n')


# noinspection PyUnusedLocal
def exit_if_idle_command(sock, args, instance):
    """Exit papa if there are no processes, sockets or values"""
    instance_globals = instance['globals']
    if not is_idle(instance_globals):
        return "not idle"
    instance_globals['exit_if_idle'] = True
    raise CloseSocket('Exiting papa!\n> ')


# noinspection PyUnusedLocal
def help_command(sock, args, instance):
    """Show help info"""
    if args:
        try:
            help_for = lookup_command(args, allow_partials=True)
        except Error as e:
            return '{0}\n'.format(e)
        return help_for['__doc__'].strip() if isinstance(help_for, dict) else help_for.__doc__
    return """Possible commands are:
    make socket - Create a socket to be used by processes
    remove sockets - Close and remove sockets by name or file number
    list sockets - List sockets by name or file number
    -----------------------------------------------------
    make process - Launch a process
    remove processes - Stop recording the output of processes by name or PID
    list processes - List processes by name or PID
    watch processes - Start receiving the output of a processes by name or PID
    -----------------------------------------------------
    set - Set a named value
    get - Get a named value
    list values - List values by name
    remove processes - Remove values by name
    -----------------------------------------------------
    quit - Close the client session
    exit-if-idle Exit papa if there are no processes, sockets or values
    help - Type "help <cmd>" for more information

NOTE: All of these commands may be abbreviated. Type at least one character of
      each word. The rest are optional.

After a 'watch' command, just enter '-' and a return to receive more output.
"""


remove_doc = """
Remove the socket, value or process output channel.

You can remove process output channels by name or PID
Examples:
    remove process 3698
    remove process nginx

You can remove sockets by name or file number
Examples:
    remove sockets uwsgi
    remove socket 10

You can remove values by name
Examples:
    remove values uwsgi.*
    remove value aack

All commands can be abbreviated as much as you like, so the above can also be:
    r process 3698
    rem proc nginx
    r sockets uwsgi
    r s 10
    r v uwsgi.*
    re val aack
"""

list_doc = """
List sockets, processes or values.

You can list processes by name or PID
Examples:
    list process 3698
    list processes nginx.*

You can list sockets by name or file number
Examples:
    list socket 10
    list sockets uwsgi.*

You can list values by name
Examples:
    list values uwsgi.*

All commands can be abbreviated as much as you like, so the above can also be:
    l process 3698
    lis proc nginx.*
    l s 10
    l sockets uwsgi.*
    li val uwsgi.*
"""

make_doc = """
Make a new socket or process.

Do 'help make process' or 'help make socket' for details.
"""

watch_doc = """
Watch processes.

You can watch processes by name or PID
Examples:
    watch processes 3698
    watch processes nginx.*

All commands can be abbreviated as much as you like, so the above can also be:
    w process 3698
    wat proc nginx.*
"""


top_level_commands = {
    'list': {
        'sockets': papa_socket.sockets_command,
        'processes': proc.processes_command,
        'values': values.values_command,
        '__doc__': list_doc
    },
    'make': {
        'socket': papa_socket.socket_command,
        'process': proc.process_command,
        '__doc__': make_doc
    },
    'remove': {
        'sockets': papa_socket.close_socket_command,
        'processes': proc.close_output_command,
        'values': values.remove_command,
        '__doc__': remove_doc
    },
    'watch': {
        'processes': proc.watch_command,
        '__doc__': watch_doc
    },
    'set': values.set_command,
    'get': values.get_command,
    'quit': quit_command,
    'exit-if-idle': exit_if_idle_command,
    'help': help_command,
}


def lookup_command(args, commands=top_level_commands, primary_command=None, allow_partials=False):
    cmd = args.pop(0).lower()
    if cmd not in commands:
        for item in sorted(commands):
            if item.startswith(cmd):
                cmd = item
                if cmd == 'exit-if-idle':
                    raise Error('You cannot abbreviate "exit-if-idle"')
                break
        else:
            if primary_command:
                raise Error('Bad "{0}" command. The following word must be one of: {1}'.format(primary_command, ', '.join(sorted(command for command in commands if not command.startswith('__')))))
            else:
                raise Error('Unknown command "{0}"'.format(cmd))

    result = commands[cmd]
    if isinstance(result, dict):
        if args:
            if primary_command:
                cmd = ' '.join((primary_command, cmd))
            return lookup_command(args, result, cmd, allow_partials)
        if not allow_partials:
            raise Error('"{0}" must be followed by one of: {1}'.format(cmd, ', '.join(sorted(command for command in result if not command.startswith('__')))))
    return result


class ServerCommandConnection(object):
    def __init__(self, sock):
        self.sock = sock
        self.data = b''

    def readline(self):
        while not b'\n' in self.data:
            new_data = recv_with_retry(self.sock)
            if not new_data:
                raise socket.error('done')
            self.data += new_data

        one_line, self.data = self.data.partition(b'\n')[::2]
        return cast_string(one_line).strip()


def chat_with_a_client(sock, addr, instance_globals, container):
    connection = ServerCommandConnection(sock)
    instance = {'globals': instance_globals, 'connection': connection}
    try:
        sock.send(b'Papa is home. Type "help" for commands.\n> ')

        while True:
            one_line = connection.readline()
            args = []
            acc = ''
            if one_line:
                for arg in one_line.split(' '):
                    if arg:
                        if arg[-1] == '\\':
                            acc += arg[:-1] + ' '
                        else:
                            acc += arg
                            args.append(acc.strip())
                            acc = ''
                if acc:
                    args.append(acc)

                try:
                    command = lookup_command(args)
                except Error as e:
                    reply = 'Error: {0}\n'.format(e)
                else:
                    try:
                        reply = command(sock, args, instance) or '\n'
                    except CloseSocket as e:
                        if e.final_message:
                            send_with_retry(sock, cast_bytes(e.final_message))
                        break
                    except papa.utils.Error as e:
                        reply = 'Error: {0}\n'.format(e)
                    except Exception as e:
                        reply = 'Error: {0}\n'.format(e)

                if reply[-1] != '\n':
                    reply += '\n> '
                else:
                    reply += '> '
                reply = cast_bytes(reply)
            else:
                reply = b'> '
            send_with_retry(sock, reply)
    except socket.error:
        pass

    try:
        sock.close()
    except socket.error:
        pass

    if container:
        thread_object = container[0]
        instance_globals['active_threads'].remove(thread_object)
        instance_globals['inactive_threads'].append((addr, thread_object))


def cleanup(instance_globals):
    if 'lock' in instance_globals:
        papa_socket.cleanup(instance_globals)


def is_idle(instance_globals):
    with instance_globals['lock']:
        return not instance_globals['processes']\
               and not instance_globals['sockets']['by_name']\
               and not instance_globals['sockets']['by_path']\
               and not instance_globals['values']


def socket_server(port_or_path, single_socket_mode=False):
    instance_globals = {
        'processes': {},
        'sockets': {'by_name': {}, 'by_path': {}},
        'values': {},
        'active_threads': [],
        'inactive_threads': [],
        'lock': Lock(),
        'exit_if_idle': False,
    }
    def local_cleanup():
        cleanup(instance_globals)
    atexit.register(local_cleanup)
    try:
        if isinstance(port_or_path, str):
            try:
                os.unlink(port_or_path)
            except OSError:
                if os.path.exists(port_or_path):
                    raise
            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            try:
                s.bind(port_or_path)
            except socket.error as e:
                raise Error('Bind failed: {0}'.format(e))
        else:
            location = ('127.0.0.1', port_or_path)
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                s.bind(location)
            except socket.error as e:
                raise Error('Bind failed or port {0}: {1}'.format(port_or_path, e))
    except Exception as e:
        log.error(e)
        sys.exit(1)

    s.listen(5)
    log.info('Listening')
    while True:
        try:
            sock, addr = s.accept()
            log.info('Started client session with %s:%d', addr[0], addr[1])
            container = []
            instance_globals['exit_if_idle'] = False
            t = Thread(target=chat_with_a_client, args=(sock, addr, instance_globals, container))
            container.append(t)
            instance_globals['active_threads'].append(t)
            t.daemon = True
            t.start()
            s.settimeout(.5)
        except socket.timeout:
            pass
        while instance_globals['inactive_threads']:
            addr, t = instance_globals['inactive_threads'].pop()
            t.join()
            log.info('Closed client session with %s:%d', addr[0], addr[1])
        if not instance_globals['active_threads']:
            if single_socket_mode:
                break
            if instance_globals['exit_if_idle'] and is_idle(instance_globals):
                log.info('Exiting due to exit_if_idle request')
                break
            s.settimeout(None)
    s.close()
    papa_socket.cleanup(instance_globals)
    try:
        # noinspection PyUnresolvedReferences
        atexit.unregister(local_cleanup)
    except AttributeError:
        del instance_globals['lock']


def daemonize_server(port_or_path, fix_title=False):
    process_id = os.fork()
    if process_id < 0:
        raise Error('Unable to fork')
    elif process_id != 0:
        return

    # noinspection PyNoneFunctionAssignment,PyArgumentList
    process_id = os.setsid()
    if process_id == -1:
        sys.exit(1)

    for fd in range(3, resource.getrlimit(resource.RLIMIT_NOFILE)[0]):
        try:
            os.close(fd)
        except OSError:
            pass

    devnull = os.devnull if hasattr(os, 'devnull') else '/dev/null'
    devnull_fd = os.open(devnull, os.O_RDWR)
    for fd in range(3):
        # noinspection PyTypeChecker
        os.dup2(devnull_fd, fd)

    os.umask(0o27)
    os.chdir('/')
    if fix_title and setproctitle is not None:
        # noinspection PyCallingNonCallable
        setproctitle('papa daemon from %s' % os.path.basename(sys.argv[0]))
    socket_server(port_or_path)


def main():
    parser = ArgumentParser('papa', description='A simple parent process for sockets and other processes')
    parser.add_argument('-d', '--debug', action='store_true', help='run in debug mode')
    parser.add_argument('-u', '--unix-socket', help='path to unix socket to bind')
    parser.add_argument('-p', '--port', default=20202, type=int, help='port to bind on localhost (default 20202)')
    parser.add_argument('--daemonize', action='store_true', help='daemonize the papa server')
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO if args.debug else logging.ERROR)
    if args.daemonize:
        daemonize_server(args.unix_socket or args.port)
    else:
        try:
            socket_server(args.unix_socket or args.port)
        except Exception as e:
            log.error(e)

if __name__ == '__main__':
    main()



import os
import sys
import os.path
import socket
from time import sleep
try:
    # noinspection PyPackageRequirements
    import unittest2 as unittest
except ImportError:
    import unittest
import select
import papa
from papa.server.papa_socket import unix_socket
from papa.utils import cast_bytes
from tempfile import gettempdir
import logging

logging.basicConfig()
import inspect

isdebugging = False
for frame in inspect.stack():
    if frame[1].endswith("pydevd.py"):
        isdebugging = True
        break

here = os.path.dirname(os.path.realpath(__file__))


class SocketTest(unittest.TestCase):
    def setUp(self):
        papa.set_debug_mode(quit_when_connection_closed=True)

    def check_subset(self, expected, result):
        for key, value in expected.items():
            self.assertIn(key, result)
            self.assertEqual(value, result[key])

    def test_inet(self):
        with papa.Papa() as p:
            expected = {'type': 'stream', 'family': 'inet', 'backlog': 5, 'host': '127.0.0.1'}
            reply = p.make_socket('inet_sock')
            self.check_subset(expected, reply)
            self.assertIn('port', reply)
            p.remove_sockets('inet_sock')
            self.assertDictEqual({}, p.list_sockets())
            reply = p.make_socket('inet_sock', reuseport=True)
            self.check_subset(expected, reply)
            self.assertIn('port', reply)

    def test_inet_interface(self):
        with papa.Papa() as p:
            expected = {'type': 'stream', 'interface': 'eth0', 'family': 'inet', 'backlog': 5, 'host': '0.0.0.0'}
            self.assertDictEqual({}, p.list_sockets())

            reply = p.make_socket('interface_socket', interface='eth0')
            self.assertIn('port', reply)
            self.assertIn('fileno', reply)
            expected['port'] = reply['port']
            expected['fileno'] = reply['fileno']
            self.assertDictEqual(expected, reply)
            self.assertDictEqual({'interface_socket': expected}, p.list_sockets())
            p.remove_sockets('interface_socket')
            self.assertDictEqual({}, p.list_sockets())
            reply = p.make_socket('interface_socket', interface='eth0', port=expected['port'])
            self.assertDictEqual(expected, reply)
            self.assertDictEqual({'interface_socket': expected}, p.list_sockets())

    def test_inet6(self):
        with papa.Papa() as p:
            expected = {'type': 'stream', 'family': 'inet6', 'backlog': 5, 'host': '::1'}
            reply = p.make_socket('inet6_sock', family=socket.AF_INET6)
            self.check_subset(expected, reply)
            self.assertIn('port', reply)
            p.remove_sockets('inet6_sock')
            self.assertDictEqual({}, p.list_sockets())
            reply = p.make_socket('inet6_sock', family=socket.AF_INET6, reuseport=True)
            self.check_subset(expected, reply)
            self.assertIn('port', reply)

    def test_inet6_interface(self):
        with papa.Papa() as p:
            expected = {'type': 'stream', 'interface': 'eth0', 'family': 'inet6', 'backlog': 5, 'host': '::'}
            self.assertDictEqual({}, p.list_sockets())

            reply = p.make_socket('interface_socket', family=socket.AF_INET6, interface='eth0')
            self.assertIn('port', reply)
            self.assertIn('fileno', reply)
            expected['port'] = reply['port']
            expected['fileno'] = reply['fileno']
            self.assertDictEqual(expected, reply)
            self.assertDictEqual({'interface_socket': expected}, p.list_sockets())

    @unittest.skipIf(unix_socket is None, 'Unix socket not supported on this platform')
    def test_file_socket(self):
        with papa.Papa() as p:
            path = os.path.join(gettempdir(), 'tst.sock')
            expected = {'path': path, 'backlog': 5, 'type': 'stream', 'family': 'unix'}

            reply = p.make_socket('fsock', path=path)
            self.assertIn('fileno', reply)
            expected['fileno'] = reply['fileno']
            self.assertDictEqual(expected, reply)
            self.assertDictEqual({'fsock': expected}, p.list_sockets())

            self.assertRaises(papa.Error, p.make_socket, 'fsock', path='path')

    def test_already_exists(self):
        with papa.Papa() as p:
            reply = p.make_socket('exists_sock')
            self.assertDictEqual(reply, p.make_socket('exists_sock'))
            self.assertRaises(papa.Error, p.make_socket, 'exists_sock', family=socket.AF_INET6)

    def test_wildcard(self):
        with papa.Papa() as p:
            expected = {'type': 'stream', 'family': 'inet', 'backlog': 5, 'host': '127.0.0.1'}

            reply = p.make_socket('inet.0')
            self.check_subset(expected, reply)
            self.assertIn('port', reply)

            reply = p.make_socket('inet.1')
            self.check_subset(expected, reply)
            self.assertIn('port', reply)

            reply = p.make_socket('other')
            self.check_subset(expected, reply)
            self.assertIn('port', reply)

            reply = p.list_sockets('inet.*')
            self.assertEqual(2, len(list(reply.keys())))
            self.assertEqual(['inet.0', 'inet.1'], sorted(reply.keys()))

            reply = p.list_sockets('other')
            self.assertEqual(1, len(list(reply.keys())))
            self.assertEqual(['other'], list(reply.keys()))

            reply = p.list_sockets('not_there')
            self.assertEqual({}, reply)

            reply = p.list_sockets('other', 'inet.1')
            self.assertEqual(2, len(list(reply.keys())))
            self.assertEqual(['inet.1', 'other'], sorted(reply.keys()))

            reply = p.list_sockets('other', 'inet*')
            self.assertEqual(3, len(list(reply.keys())))
            self.assertEqual(['inet.0', 'inet.1', 'other'], sorted(reply.keys()))

            reply = p.list_sockets('*')
            self.assertEqual(3, len(list(reply.keys())))
            self.assertEqual(['inet.0', 'inet.1', 'other'], sorted(reply.keys()))

            reply = p.list_sockets()
            self.assertEqual(3, len(list(reply.keys())))
            self.assertEqual(['inet.0', 'inet.1', 'other'], sorted(reply.keys()))


class ValueTest(unittest.TestCase):
    def setUp(self):
        papa.set_debug_mode(quit_when_connection_closed=True)

    def test_value(self):
        with papa.Papa() as p:
            self.assertEqual(None, p.get('aack'))
            self.assertDictEqual({}, p.list_values())

            p.set('aack', 'bar')
            self.assertEqual('bar', p.get('aack'))
            self.assertDictEqual({'aack': 'bar'}, p.list_values())

            p.set('aack2', 'barry')
            self.assertEqual('barry', p.get('aack2'))
            self.assertDictEqual({'aack': 'bar', 'aack2': 'barry'}, p.list_values())

            p.set('aack3', 'larry')
            self.assertEqual('larry', p.get('aack3'))
            self.assertDictEqual({'aack': 'bar', 'aack2': 'barry', 'aack3': 'larry'}, p.list_values())

            p.set('bar', 'aack')
            self.assertEqual('aack', p.get('bar'))
            self.assertDictEqual({'aack': 'bar', 'aack2': 'barry', 'aack3': 'larry', 'bar': 'aack'}, p.list_values())

            self.assertDictEqual({'aack': 'bar', 'aack2': 'barry', 'aack3': 'larry', 'bar': 'aack'}, p.list_values('*'))
            self.assertDictEqual({'aack': 'bar', 'aack2': 'barry', 'aack3': 'larry'}, p.list_values('aack*'))
            self.assertDictEqual({'bar': 'aack'}, p.list_values('b*'))
            self.assertDictEqual({'aack2': 'barry', 'bar': 'aack'}, p.list_values('aack2', 'b*'))

            p.set('aack')
            self.assertEqual(None, p.get('aack'))
            self.assertDictEqual({'aack2': 'barry', 'aack3': 'larry'}, p.list_values('a*'))

            p.remove_values('aack*')
            self.assertDictEqual({'bar': 'aack'}, p.list_values())

    def test_wildcard_clear(self):
        with papa.Papa() as p:
            self.assertRaises(papa.Error, p.remove_values)
            self.assertRaises(papa.Error, p.remove_values, '*')


class ProcessTest(unittest.TestCase):
    def setUp(self):
        papa.set_debug_mode(quit_when_connection_closed=True)

    @staticmethod
    def _merge_lines(output):
        if len(output) > 1:
            by_name = {}
            merged_lines = False
            for line in output:
                by_name.setdefault(line.name, []).append(line)
            for named_lines in by_name.values():
                line_number = 1
                while line_number < len(named_lines):
                    line = named_lines[line_number]
                    prev = named_lines[line_number - 1]
                    if line.timestamp - prev.timestamp > .05:
                        line_number += 1
                    else:
                        named_lines[line_number - 1] = papa.ProcessOutput(prev.name, prev.timestamp, prev.data + line.data)
                        del named_lines[line_number]
                        merged_lines = True
            if merged_lines:
                output = sorted((item for items in by_name.values() for item in items), key=lambda x: x.timestamp)
        return output

    def gather_output(self, watcher):
        out = []
        err = []
        close = []
        while watcher:
            reply = watcher.read()
            if reply:
                out.extend(reply[0])
                err.extend(reply[1])
                close.extend(reply[2])
        return self._merge_lines(out), self._merge_lines(err), close

    if isdebugging:
        _non_debug_gather_output = gather_output

        def _filter_list(self, output):
            remove = []
            for line_number, line in enumerate(output):
                if line.data.startswith(b'pydev debugger: process ') and b'is connecting' in line.data:
                    if line.data.endswith(b'is connecting\n\n'):
                        remove.append(line_number)
                    else:
                        end = line.data.find(b'is connecting\n\n')
                        output[line_number] = papa.ProcessOutput(line.name, line.timestamp, line.data[end + 14:])
            for line_number in reversed(remove):
                del output[line_number]

        def gather_output(self, watcher):
            out, err, close = self._non_debug_gather_output(watcher)
            self._filter_list(out)
            self._filter_list(err)
            return out, err, close

    def test_process_with_out_and_err(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            reply1 = p.make_process('write3', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertIn('pid', reply1)
            self.assertIsInstance(reply1['pid'], int)
            reply = p.list_processes()
            self.assertEqual(1, len(list(reply.keys())))
            self.assertEqual('write3', list(reply.keys())[0])
            self.assertIn('pid', list(reply.values())[0])

            reply2 = p.make_process('write3', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertEqual(reply1['pid'], reply2['pid'])

            self.assertRaises(papa.Error, p.watch_processes, 'not_there')

            with p.watch_processes('write*') as w:
                select.select([w], [], [])
                self.assertTrue(w.ready)
                out, err, close = self.gather_output(w)
                exit_code = w.exit_code['write3']
            self.assertEqual(3, len(out))
            self.assertEqual(1, len(err))
            self.assertEqual(1, len(close))
            self.assertEqual('write3', out[0].name)
            self.assertEqual('write3', out[1].name)
            self.assertEqual('write3', out[2].name)
            self.assertEqual('write3', err[0].name)
            self.assertEqual('write3', close[0].name)
            self.assertLess(out[0].timestamp, out[1].timestamp)
            self.assertLess(out[1].timestamp, out[2].timestamp)
            self.assertLessEqual(out[2].timestamp, err[0].timestamp)
            self.assertLessEqual(out[2].timestamp, close[0].timestamp)
            self.assertLessEqual(err[0].timestamp, close[0].timestamp)
            self.assertEqual(b'Version: ' + cast_bytes(sys.version.partition(' ')[0]) + b'\n', out[0].data)
            self.assertEqual(b'Executable: ' + cast_bytes(sys.executable) + b'\n', out[1].data)
            self.assertEqual(b'Args: \n', out[2].data)
            self.assertEqual(b'done', err[0].data)
            self.assertEqual(0, close[0].data)
            self.assertEqual(0, exit_code)
            self.assertDictEqual({}, p.list_processes())

    def test_process_with_none_executable(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            reply1 = p.make_process('write3', None, args=[sys.executable, 'executables/write_three_lines.py'], working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertIn('pid', reply1)
            self.assertIsInstance(reply1['pid'], int)
            reply = p.list_processes()
            self.assertEqual(1, len(list(reply.keys())))
            self.assertEqual('write3', list(reply.keys())[0])
            self.assertIn('pid', list(reply.values())[0])

            reply2 = p.make_process('write3', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertEqual(reply1['pid'], reply2['pid'])

            self.assertRaises(papa.Error, p.watch_processes, 'not_there')

            with p.watch_processes('write*') as w:
                select.select([w], [], [])
                self.assertTrue(w.ready)
                out, err, close = self.gather_output(w)
                exit_code = w.exit_code['write3']
            self.assertEqual(3, len(out))
            self.assertEqual(1, len(err))
            self.assertEqual(1, len(close))
            self.assertEqual('write3', out[0].name)
            self.assertEqual('write3', out[1].name)
            self.assertEqual('write3', out[2].name)
            self.assertEqual('write3', err[0].name)
            self.assertEqual('write3', close[0].name)
            self.assertLess(out[0].timestamp, out[1].timestamp)
            self.assertLess(out[1].timestamp, out[2].timestamp)
            self.assertLessEqual(out[2].timestamp, err[0].timestamp)
            self.assertLessEqual(out[2].timestamp, close[0].timestamp)
            self.assertLessEqual(err[0].timestamp, close[0].timestamp)
            self.assertEqual(b'Version: ' + cast_bytes(sys.version.partition(' ')[0]) + b'\n', out[0].data)
            self.assertEqual(b'Executable: ' + cast_bytes(sys.executable) + b'\n', out[1].data)
            self.assertEqual(b'Args: \n', out[2].data)
            self.assertEqual(b'done', err[0].data)
            self.assertEqual(0, close[0].data)
            self.assertEqual(0, exit_code)
            self.assertDictEqual({}, p.list_processes())

    def test_process_with_watch_immediately(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            with p.make_process('write3', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ, watch_immediately=True) as w:
                out, err, close = self.gather_output(w)
                exit_code = w.exit_code['write3']
            self.assertEqual(3, len(out))
            self.assertEqual(1, len(err))
            self.assertEqual(1, len(close))
            self.assertEqual('write3', out[0].name)
            self.assertEqual('write3', out[1].name)
            self.assertEqual('write3', out[2].name)
            self.assertEqual('write3', err[0].name)
            self.assertEqual('write3', close[0].name)
            self.assertLess(out[0].timestamp, out[1].timestamp)
            self.assertLess(out[1].timestamp, out[2].timestamp)
            self.assertLessEqual(out[2].timestamp, err[0].timestamp)
            self.assertLessEqual(out[2].timestamp, close[0].timestamp)
            self.assertLessEqual(err[0].timestamp, close[0].timestamp)
            self.assertEqual(b'Version: ' + cast_bytes(sys.version.partition(' ')[0]) + b'\n', out[0].data)
            self.assertEqual(b'Executable: ' + cast_bytes(sys.executable) + b'\n', out[1].data)
            self.assertEqual(b'Args: \n', out[2].data)
            self.assertEqual(b'done', err[0].data)
            self.assertEqual(0, close[0].data)
            self.assertEqual(0, exit_code)
            self.assertDictEqual({}, p.list_processes())

    def test_process_with_err_redirected_to_out(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            reply1 = p.make_process('write3', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ, stderr=papa.STDOUT)
            self.assertIn('pid', reply1)
            self.assertIsInstance(reply1['pid'], int)
            reply = p.list_processes()
            self.assertEqual(1, len(list(reply.keys())))
            self.assertEqual('write3', list(reply.keys())[0])
            self.assertIn('pid', list(reply.values())[0])

            with p.watch_processes('write*') as w:
                out, err, close = self.gather_output(w)
                exit_code = w.exit_code['write3']
            self.assertEqual(3, len(out))
            self.assertEqual(0, len(err))
            self.assertEqual(1, len(close))
            self.assertEqual('write3', out[0].name)
            self.assertEqual('write3', out[1].name)
            self.assertEqual('write3', out[2].name)
            self.assertEqual('write3', close[0].name)
            self.assertLess(out[0].timestamp, out[1].timestamp)
            self.assertLess(out[1].timestamp, out[2].timestamp)
            self.assertLessEqual(out[2].timestamp, close[0].timestamp)
            self.assertEqual(b'Version: ' + cast_bytes(sys.version.partition(' ')[0]) + b'\n', out[0].data)
            self.assertEqual(b'Executable: ' + cast_bytes(sys.executable) + b'\n', out[1].data)
            self.assertEqual(b'Args: \ndone', out[2].data)
            self.assertEqual(0, close[0].data)
            self.assertEqual(0, exit_code)
            self.assertDictEqual({}, p.list_processes())

    def test_process_with_no_out(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            reply1 = p.make_process('write3', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ, stdout=papa.DEVNULL)
            self.assertIn('pid', reply1)
            self.assertIsInstance(reply1['pid'], int)
            reply = p.list_processes()
            self.assertEqual(1, len(list(reply.keys())))
            self.assertEqual('write3', list(reply.keys())[0])
            self.assertIn('pid', list(reply.values())[0])

            with p.watch_processes('write*') as w:
                out, err, close = self.gather_output(w)
                exit_code = w.exit_code['write3']
            self.assertEqual(0, len(out))
            self.assertEqual(1, len(err))
            self.assertEqual(1, len(close))
            self.assertEqual('write3', err[0].name)
            self.assertEqual('write3', close[0].name)
            self.assertLessEqual(err[0].timestamp, close[0].timestamp)
            self.assertEqual(b'done', err[0].data)
            self.assertEqual(0, close[0].data)
            self.assertEqual(0, exit_code)
            self.assertDictEqual({}, p.list_processes())

    def test_process_with_no_buffer(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            reply1 = p.make_process('write3', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ, bufsize=0)
            self.assertIn('pid', reply1)
            self.assertIsInstance(reply1['pid'], int)
            reply = p.list_processes()
            self.assertEqual(1, len(list(reply.keys())))
            self.assertEqual('write3', list(reply.keys())[0])
            self.assertIn('pid', list(reply.values())[0])

            with p.watch_processes('write*') as w:
                out, err, close = self.gather_output(w)
                exit_code = w.exit_code['write3']
            self.assertEqual(0, len(out))
            self.assertEqual(0, len(err))
            self.assertEqual(1, len(close))
            self.assertEqual('write3', close[0].name)
            self.assertEqual(0, close[0].data)
            self.assertEqual(0, exit_code)
            self.assertDictEqual({}, p.list_processes())

    def test_two_list_processes_full_output(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            reply1 = p.make_process('write3.0', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertIn('pid', reply1)
            self.assertIsInstance(reply1['pid'], int)

            reply2 = p.make_process('write3.1', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertIn('pid', reply2)
            self.assertIsInstance(reply2['pid'], int)

            reply = p.list_processes()
            self.assertEqual(2, len(list(reply.keys())))
            self.assertEqual(['write3.0', 'write3.1'], sorted(reply.keys()))
            self.assertIn('pid', list(reply.values())[0])
            self.assertIn('pid', list(reply.values())[1])
            self.assertNotEqual(list(reply.values())[0]['pid'], list(reply.values())[1]['pid'])

            with p.watch_processes('write3.*') as w:
                select.select([w], [], [])
                self.assertTrue(w.ready)
                out, err, close = self.gather_output(w)
                exit_code0 = w.exit_code['write3.0']
                exit_code1 = w.exit_code['write3.1']
            self.assertEqual(6, len(out))
            self.assertEqual(2, len(err))
            self.assertEqual(2, len(close))
            self.assertEqual(3, len([item for item in out if item.name == 'write3.0']))
            self.assertEqual(3, len([item for item in out if item.name == 'write3.1']))
            self.assertEqual(1, len([item for item in err if item.name == 'write3.0']))
            self.assertEqual(1, len([item for item in err if item.name == 'write3.1']))
            self.assertEqual(1, len([item for item in close if item.name == 'write3.0']))
            self.assertEqual(1, len([item for item in close if item.name == 'write3.1']))
            self.assertEqual(b'done', err[0].data)
            self.assertEqual(b'done', err[1].data)
            self.assertEqual(0, close[0].data)
            self.assertEqual(0, exit_code0)
            self.assertEqual(0, close[1].data)
            self.assertEqual(0, exit_code1)
            self.assertDictEqual({}, p.list_processes())

    def test_two_list_processes_wait_for_one_to_close(self):
        with papa.Papa() as p:
            f = p.fileno()
            self.assertDictEqual({}, p.list_processes())
            reply1 = p.make_process('write3.0', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertIn('pid', reply1)
            self.assertIsInstance(reply1['pid'], int)

            sleep(.2)
            reply2 = p.make_process('write3.1', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertIn('pid', reply2)
            self.assertIsInstance(reply2['pid'], int)

            reply = p.list_processes()
            self.assertEqual(2, len(list(reply.keys())))
            self.assertEqual(['write3.0', 'write3.1'], sorted(reply.keys()))
            self.assertIn('pid', list(reply.values())[0])
            self.assertIn('pid', list(reply.values())[1])
            self.assertNotEqual(list(reply.values())[0]['pid'], list(reply.values())[1]['pid'])

            with p.watch_processes('write3.*') as w:
                while True:
                    out, err, close = w.read()
                    if close:
                        break
            reply = p.list_processes()
            self.assertEqual(1, len(list(reply.keys())))
            self.assertEqual('write3.1', list(reply.keys())[0])
            self.assertEqual(f, p.fileno())

    def test_multiple_watchers(self):
        with papa.Papa() as p:
            f = p.fileno()
            self.assertDictEqual({}, p.list_processes())
            reply1 = p.make_process('write3.0', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertIn('pid', reply1)
            self.assertIsInstance(reply1['pid'], int)

            reply2 = p.make_process('write3.1', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertIn('pid', reply2)
            self.assertIsInstance(reply2['pid'], int)

            reply = p.list_processes()
            self.assertEqual(2, len(list(reply.keys())))
            self.assertEqual(['write3.0', 'write3.1'], sorted(reply.keys()))
            self.assertIn('pid', list(reply.values())[0])
            self.assertIn('pid', list(reply.values())[1])
            self.assertNotEqual(list(reply.values())[0]['pid'], list(reply.values())[1]['pid'])

            w1 = p.watch_processes('write3.0')
            self.assertEqual(f, w1.fileno())

            w2 = p.watch_processes('write3.1')
            self.assertNotEqual(f, w2.fileno())

            p.set('p1', 't1')
            self.assertNotEqual(f, p.fileno())
            self.assertNotEqual(p.fileno(), w1.fileno())
            self.assertNotEqual(p.fileno(), w2.fileno())

            out1, err1, close1 = self.gather_output(w1)
            out2, err2, close2 = self.gather_output(w2)

            w1.close()
            w2.close()

            self.assertEqual(3, len(out1))
            self.assertEqual(1, len(err1))
            self.assertEqual(1, len(close1))
            self.assertEqual('write3.0', out1[0].name)
            self.assertEqual('write3.0', out1[1].name)
            self.assertEqual('write3.0', out1[2].name)
            self.assertEqual('write3.0', err1[0].name)
            self.assertEqual('write3.0', close1[0].name)
            self.assertLess(out1[0].timestamp, out1[1].timestamp)
            self.assertLess(out1[1].timestamp, out1[2].timestamp)
            self.assertLessEqual(out1[2].timestamp, err1[0].timestamp)
            self.assertLessEqual(out1[2].timestamp, close1[0].timestamp)
            self.assertLessEqual(err1[0].timestamp, close1[0].timestamp)
            self.assertEqual(b'Version: ' + cast_bytes(sys.version.partition(' ')[0]) + b'\n', out1[0].data)
            self.assertEqual(b'Executable: ' + cast_bytes(sys.executable) + b'\n', out1[1].data)
            self.assertEqual(b'Args: \n', out1[2].data)
            self.assertEqual(b'done', err1[0].data)
            self.assertEqual(0, close1[0].data)

            self.assertEqual(3, len(out2))
            self.assertEqual(1, len(err2))
            self.assertEqual(1, len(close2))
            self.assertEqual('write3.1', out2[0].name)
            self.assertEqual('write3.1', out2[1].name)
            self.assertEqual('write3.1', out2[2].name)
            self.assertEqual('write3.1', err2[0].name)
            self.assertEqual('write3.1', close2[0].name)
            self.assertLess(out2[0].timestamp, out2[1].timestamp)
            self.assertLess(out2[1].timestamp, out2[2].timestamp)
            self.assertLessEqual(out2[2].timestamp, err2[0].timestamp)
            self.assertLessEqual(out2[2].timestamp, close2[0].timestamp)
            self.assertLessEqual(err2[0].timestamp, close2[0].timestamp)
            self.assertEqual(b'Version: ' + cast_bytes(sys.version.partition(' ')[0]) + b'\n', out2[0].data)
            self.assertEqual(b'Executable: ' + cast_bytes(sys.executable) + b'\n', out2[1].data)
            self.assertEqual(b'Args: \n', out2[2].data)
            self.assertEqual(b'done', err2[0].data)
            self.assertEqual(0, close2[0].data)
            self.assertDictEqual({}, p.list_processes())

            self.assertEqual('t1', p.get('p1'))

    def test_process_with_small_buffer(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            reply1 = p.make_process('write3', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ, bufsize=14)
            self.assertIn('pid', reply1)
            self.assertIsInstance(reply1['pid'], int)
            reply = p.list_processes()
            self.assertEqual(1, len(list(reply.keys())))
            self.assertEqual('write3', list(reply.keys())[0])
            self.assertIn('pid', list(reply.values())[0])
            sleep(.7)

            with p.watch_processes('write*') as w:
                select.select([w], [], [])
                self.assertTrue(w.ready)
                out, err, close = self.gather_output(w)
                exit_code = w.exit_code['write3']
            self.assertEqual(1, len(out))
            self.assertEqual(1, len(err))
            self.assertEqual(1, len(close))
            self.assertEqual('write3', out[0].name)
            self.assertEqual('write3', err[0].name)
            self.assertEqual('write3', close[0].name)
            self.assertLessEqual(out[0].timestamp, err[0].timestamp)
            self.assertLessEqual(out[0].timestamp, close[0].timestamp)
            self.assertLessEqual(err[0].timestamp, close[0].timestamp)
            self.assertEqual(b'Args: \n', out[0].data)
            self.assertEqual(b'done', err[0].data)
            self.assertEqual(0, close[0].data)
            self.assertEqual(0, exit_code)
            self.assertDictEqual({}, p.list_processes())

    def test_one_process_two_parallel_watchers(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            reply = p.make_process('write3', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertIn('pid', reply)
            self.assertIn('running', reply)
            self.assertIsInstance(reply['pid'], int)
            self.assertIsInstance(reply['running'], bool)

            w1 = p.watch_processes('write*')
            w2 = p.watch_processes('write*')
            out1, err1, close1 = w1.read()
            if isdebugging and not out1 and err1 and err1[0].data.startswith('pydev debugger'):
                out1, err1, close1 = w1.read()
            out2, err2, close2 = w2.read()
            self.assertEqual(out1[0], out2[0])
            w1.close()
            w2.close()

    def test_one_process_two_serial_watchers(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            reply = p.make_process('write3', sys.executable, args='executables/write_three_lines.py', working_dir=here, uid=os.environ['LOGNAME'], env=os.environ)
            self.assertIn('pid', reply)
            self.assertIn('running', reply)
            self.assertIsInstance(reply['pid'], int)
            self.assertIsInstance(reply['running'], bool)

            with p.watch_processes('write*') as w:
                out1, err1, close1 = w.read()
                if isdebugging and not out1 and err1 and err1[0].data.startswith('pydev debugger'):
                    out1, err1, close1 = w.read()
            with p.watch_processes('write*') as w:
                out2, err2, close2 = self.gather_output(w)
            self.assertLess(out1[0].timestamp, out2[0].timestamp)

    def test_echo_server_with_normal_socket(self):
        with papa.Papa() as p:
            reply = p.make_socket('echo_socket')
            self.assertIn('port', reply)
            self.assertIn('fileno', reply)
            port = reply['port']

            reply = p.make_process('echo1', sys.executable, args=('executables/echo_server.py', '$(socket.echo_socket.fileno)'), working_dir=here)
            self.assertIn('pid', reply)

            s = socket.socket()
            s.connect(('127.0.0.1', port))

            s.send(b'test\n')
            msg = b''
            while len(msg) < 5:
                msg += s.recv(5)
            self.assertEqual(b'test\n', msg)

            s.send(b'and do some more\n')
            msg = b''
            while len(msg) < 17:
                msg += s.recv(17)
            self.assertEqual(b'and do some more\n', msg)

            s.close()
            with p.watch_processes('echo*') as w:
                out, err, close = self.gather_output(w)
            self.assertEqual(b'test\nand do some more\n', out[0].data)

    def test_echo_server_with_echo_client(self):
        with papa.Papa() as p:
            reply = p.make_socket('echo_socket')
            self.assertIn('port', reply)
            self.assertIn('fileno', reply)

            reply = p.make_process('echo.server', sys.executable, args=('executables/echo_server.py', '$(socket.echo_socket.fileno)'), working_dir=here)
            self.assertIn('pid', reply)

            reply = p.make_process('echo.client', sys.executable, args=('executables/echo_client.py', '$(socket.echo_socket.port)'), working_dir=here)
            self.assertIn('pid', reply)

            with p.watch_processes('echo.*') as w:
                out, err, close = self.gather_output(w)
            self.assertEqual(2, len(out))
            self.assertEqual(2, len(close))
            self.assertEqual(b'howdy\n', out[0].data)
            self.assertEqual(b'howdy\n', out[1].data)
            self.assertIn(out[0].name, ('echo.client', 'echo.server'))
            self.assertIn(out[1].name, ('echo.client', 'echo.server'))
            self.assertNotEqual(out[0].name, out[1].name)

    def test_echo_server_with_reuseport(self):
        with papa.Papa() as p:
            reply = p.make_socket('echo_socket', reuseport=True)
            self.assertIn('port', reply)
            port = reply['port']

            reply = p.make_process('echo1', sys.executable, args=('executables/echo_server.py', '$(socket.echo_socket.fileno)'), working_dir=here)
            self.assertIn('pid', reply)

            s = socket.socket()
            s.connect(('127.0.0.1', port))

            s.send(b'test\n')
            msg = b''
            while len(msg) < 5:
                msg += s.recv(5)
            self.assertEqual(b'test\n', msg)

            s.send(b'and do some more\n')
            msg = b''
            while len(msg) < 17:
                msg += s.recv(17)
            self.assertEqual(b'and do some more\n', msg)

            s.close()
            with p.watch_processes('echo*') as w:
                out, err, close = self.gather_output(w)
            self.assertEqual(b'test\nand do some more\n', out[0].data)

    def test_process_with_close_output_late(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            socket_reply = p.make_socket('echo_socket')
            p.make_process('echo', sys.executable, args=('executables/echo_server.py', '$(socket.echo_socket.fileno)'), working_dir=here)

            s = socket.socket()
            s.connect(('127.0.0.1', socket_reply['port']))

            s.send(b'test\n')
            msg = b''
            while len(msg) < 5:
                msg += s.recv(5)
            self.assertEqual(b'test\n', msg)
            s.close()

            sleep(.4)
            p.remove_processes('echo')
            self.assertDictEqual({}, p.list_processes())

    def test_process_with_close_output_early(self):
        with papa.Papa() as p:
            self.assertDictEqual({}, p.list_processes())
            socket_reply = p.make_socket('echo_socket')
            p.make_process('echo', sys.executable, args=('executables/echo_server.py', '$(socket.echo_socket.fileno)'), working_dir=here)
            p.remove_processes('echo')

            s = socket.socket()
            s.connect(('127.0.0.1', socket_reply['port']))

            s.send(b'test\n')
            msg = b''
            while len(msg) < 5:
                msg += s.recv(5)
            self.assertEqual(b'test\n', msg)
            s.close()

            sleep(.4)
            self.assertDictEqual({}, p.list_processes())

    def test_bad_socket_reference(self):
        with papa.Papa() as p:
            self.assertRaises(papa.Error, p.make_process, 'bad', sys.executable, args=('executables/echo_server.py', '$(socket.echo_socket.fileno)'), working_dir=here)

    def test_bad_process(self):
        with papa.Papa() as p:
            self.assertRaises(papa.Error, p.make_process, 'bad', sys.executable + '-blah')

    def test_bad_working_dir(self):
        with papa.Papa() as p:
            self.assertRaises(papa.Error, p.make_process, 'bad', sys.executable, working_dir=here + '-blah')

if __name__ == '__main__':
    unittest.main()



import os.path
import socket
from threading import Lock
from time import sleep
from subprocess import PIPE, STDOUT
from collections import namedtuple
import logging
import select
from papa.utils import string_type, recv_with_retry, send_with_retry

try:
    from subprocess import DEVNULL
except ImportError:
    DEVNULL = -3

__author__ = 'Scott Maxwell'
__all__ = ['Papa', 'DEBUG_MODE_NONE', 'DEBUG_MODE_THREAD', 'DEBUG_MODE_PROCESS']

log = logging.getLogger('papa.client')
ProcessOutput = namedtuple('ProcessOutput', 'name timestamp data')


def wrap_trailing_slash(value):
    value = str(value)
    if value[-1] == '\\':
        value = '"{0}"'.format(value)
    return value


def append_if_not_none(container, **kwargs):
    for key, value in kwargs.items():
        if value is not None:
            value = wrap_trailing_slash(value)
            container.append('{0}={1}'.format(key, value))


class Watcher(object):
    def __init__(self, papa_object):
        self.papa_object = papa_object
        self.connection = papa_object.connection
        self.exit_code = {}
        self._fileno = self.connection.sock.fileno()
        self._need_ack = False

    def __enter__(self):
        return self

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __bool__(self):
        return self.connection is not None

    def __len__(self):
        return 1 if self.connection is not None else 0

    def fileno(self):
        return self.connection.sock.fileno()

    @property
    def ready(self):
        try:
            select.select([self], [], [], 0)
            return True
        except Exception:
            return False

    def read(self):
        self.acknowledge()
        reply = {'out': [], 'err': [], 'closed': []}
        if self.connection:
            line = b''
            while True:
                line = self.connection.get_one_line_response(b'] ')
                split = line.split(':')
                if len(split) < 4:
                    break
                result_type, name, timestamp, data = split
                data = int(data)
                if result_type == 'closed':
                    self.exit_code[name] = data
                else:
                    data = self.connection.read_bytes(data + 1)[:-1]
                result = ProcessOutput(name, float(timestamp), data)
                reply[result_type].append(result)
            self._need_ack = line == '] '
            if not self._need_ack:
                self.connection.read_bytes(2)
                if not self.papa_object.connection:
                    self.papa_object.connection = self.connection
                else:
                    self.connection.close()
                self.connection = None
                return None
        return reply['out'], reply['err'], reply['closed']

    def acknowledge(self):
        if self._need_ack:
            send_with_retry(self.connection.sock, b'\n')
            self._need_ack = False

    def close(self):
        if self.connection:
            # if the server is waiting for an ack, we can
            if self._need_ack:
                send_with_retry(self.connection.sock, b'q\n')
                self._need_ack = False
                self.connection.get_full_response()

                # we can only recover the connection if we were able to send
                # the quit ack. otherwise close the connection and let the
                # socket die
                if not self.papa_object.connection:
                    self.papa_object.connection = self.connection
                else:
                    self.connection.close()
            else:
                self.connection.close()
            self.connection = None


class ClientCommandConnection(object):
    def __init__(self, family, location):
        self.family = family
        self.location = location
        self.sock = self._attempt_to_connect()
        self.data = b''

    def _attempt_to_connect(self):
        # Try to connect to an existing Papa
        sock = socket.socket(self.family, socket.SOCK_STREAM)
        if self.family == socket.AF_INET:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.connect(self.location)
            return sock
        except Exception:
            sock.close()
            raise

    def send_command(self, command):
        if isinstance(command, list):
            command = ' '.join(c.replace(' ', '\ ').replace('\n', '\ ') for c in command if c)
        command = b(command)
        send_with_retry(self.sock, command + b'\n')

    def do_command(self, command):
        self.send_command(command)
        return self.get_full_response()

    def get_full_response(self):
        data = self.data
        self.data = b''
        while not data.endswith(b'\n> '):
            new_data = recv_with_retry(self.sock)
            if not new_data:
                raise utils.Error('Lost connection')
            data += new_data

        data = s(data[:-3])
        # noinspection PyTypeChecker
        if data.startswith('Error:'):
            raise utils.Error(data[7:])
        return data

    def get_one_line_response(self, alternate_terminator=None):
        data = self.data
        while b'\n' not in data:
            if alternate_terminator and data.endswith(alternate_terminator):
                break
            new_data = recv_with_retry(self.sock)
            if not new_data:
                raise utils.Error('Lost connection')
            data += new_data

        if data.startswith(b'Error:'):
            self.data = data
            return self.get_full_response()

        data, self.data = data.partition(b'\n')[::2]
        return s(data)

    def read_bytes(self, size):
        data = self.data
        while len(data) < size:
            new_data = recv_with_retry(self.sock, size - len(data))
            if not new_data:
                raise utils.Error('Lost connection')
            data += new_data

        data, self.data = data[:size], data[size:]
        return data

    def push_newline(self):
        self.data = b'\n' + self.data

    def close(self):
        return self.sock.close()


class Papa(object):
    _debug_mode = False
    _single_connection_mode = False
    _default_port_or_path = 20202

    spawn_lock = Lock()
    spawned = False

    def __init__(self, port_or_path=None):
        port_or_path = port_or_path or self._default_port_or_path
        self.port_or_path = port_or_path
        if isinstance(port_or_path, str):
            if not hasattr(socket, 'AF_UNIX'):
                raise NotImplementedError('Unix sockets are not supported on'
                                          ' this platform')
            if not os.path.isabs(port_or_path):
                raise utils.Error('Path to Unix socket must be absolute.')
            self.family = socket.AF_UNIX
            self.location = port_or_path
        else:
            self.family = socket.AF_INET
            self.location = ('127.0.0.1', port_or_path)

        # Try to connect to an existing Papa
        self.connection = None
        self.t = None
        self._connect(True)

    def __enter__(self):
        return self

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if self._single_connection_mode and self.t and not exc_type:
            self.t.join()
            Papa.spawned = False

    def _connect(self, allow_papa_spawn=False):
        try:
            self.connection = ClientCommandConnection(self.family, self.location)
            self.connection.get_full_response()
        except Exception:
            for i in range(100):
                if allow_papa_spawn and not Papa.spawned:
                    self._spawn_papa_server()
                try:
                    self.connection = ClientCommandConnection(self.family, self.location)
                    self.connection.get_full_response()
                    break
                except Exception:
                    sleep(.1)
            if not self.connection:
                log.error('Could not connect to Papa in 10 seconds')
                raise utils.Error('Could not connect to Papa in 10 seconds')

    def _spawn_papa_server(self):
        with Papa.spawn_lock:
            if not Papa.spawned:
                if self._debug_mode:
                    from papa.server import socket_server
                    from threading import Thread
                    t = Thread(target=socket_server, args=(self.port_or_path, self._single_connection_mode))
                    t.daemon = True
                    t.start()
                    self.t = t
                else:
                    from papa.server import daemonize_server
                    log.info('Daemonizing Papa')
                    daemonize_server(self.port_or_path, fix_title=True)
                Papa.spawned = True

    def _attempt_to_connect(self):
        # Try to connect to an existing Papa
        sock = socket.socket(self.family, socket.SOCK_STREAM)
        if self.family == socket.AF_INET:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.connect(self.location)
        return sock

    def _send_command(self, command):
        if not self.connection:
            self._connect()
        self.connection.send_command(command)

    def _do_command(self, command):
        if not self.connection:
            self._connect()
        return self.connection.do_command(command)

    @staticmethod
    def _make_socket_dict(socket_info):
        name, args = socket_info.partition(' ')[::2]
        args = dict(item.partition('=')[::2] for item in args.split(' '))
        for key in ('backlog', 'port', 'fileno'):
            if key in args:
                args[key] = int(args[key])
        return name, args

    def fileno(self):
        return self.connection.sock.fileno() if self.connection else None

    def list_sockets(self, *args):
        result = self._do_command(['l', 's'] + list(args))
        if not result:
            return {}
        # noinspection PyTypeChecker
        return dict(self._make_socket_dict(item) for item in result.split('\n'))

    def make_socket(self, name, host=None, port=None,
                    family=None, socket_type=None,
                    backlog=None, path=None, umask=None,
                    interface=None, reuseport=None):
        if not name:
            raise utils.Error('Socket requires a name')
        command = ['m', 's', name]
        if family is not None:
            try:
                family_name = utils.valid_families_by_number[family]
                command.append('family={0}'.format(family_name))
            except KeyError:
                raise utils.Error('Invalid socket family')
        if socket_type is not None:
            try:
                type_name = utils.valid_types_by_number[socket_type]
                command.append('type={0}'.format(type_name))
            except KeyError:
                raise utils.Error('Invalid socket type')
        append_if_not_none(command, backlog=backlog)
        if path:
            if not path[0] == '/' or path[-1] in '/\\':
                raise utils.Error('Socket path must be absolute to a file')
            command.append('path={0}'.format(path))
            append_if_not_none(command, umask=umask)
        else:
            append_if_not_none(command, host=host, port=port, interface=interface)
            if reuseport:
                command.append('reuseport=1')
        return self._make_socket_dict(self._do_command(command))[1]

    def remove_sockets(self, *args):
        self._do_command(['r', 's'] + list(args))
        return True

    def list_values(self, *args):
        result = self._do_command(['l', 'v'] + list(args))
        if not result:
            return {}
        # noinspection PyTypeChecker
        return dict(item.partition(' ')[::2] for item in result.split('\n'))

    def set(self, name, value=None):
        command = ['set', name]
        if value:
            command.append(value)
        self._do_command(command)

    def get(self, name):
        result = self._do_command(['get', name])
        return result or None  # do it this way so that '' becomes None

    def remove_values(self, *args):
        self._do_command(['r', 'v'] + list(args))
        return True

    @staticmethod
    def _make_process_dict(socket_info):
        name, arg_string = socket_info.partition(' ')[::2]
        args = {}
        last_key = None
        for item in arg_string.split(' '):
            key, delim, value = item.partition('=')
            if not delim and last_key:
                args[last_key] += ' ' + key
            else:
                last_key = key
                if key == 'pid':
                    value = int(value)
                elif key == 'started':
                    value = float(value)
                elif key in ('running', 'shell'):
                    value = value == 'True'
                args[key] = value
        return name, args

    def list_processes(self, *args):
        result = self._do_command(['l', 'p'] + list(args))
        if not result:
            return {}
        # noinspection PyTypeChecker
        return dict(self._make_process_dict(item) for item in result.split('\n'))

    def make_process(self, name, executable=None, args=None, env=None, working_dir=None, uid=None, gid=None, rlimits=None, stdout=None, stderr=None, bufsize=None, watch_immediately=None):
        command = ['m', 'p', name]
        append_if_not_none(command, working_dir=working_dir, uid=uid, gid=gid, bufsize=bufsize)
        if watch_immediately:
            command.append('watch=1')
        if bufsize != 0:
            if stdout is not None:
                if stdout == DEVNULL:
                    command.append('stdout=0')
                elif stdout != PIPE:
                    raise utils.Error('stdout must be DEVNULL or PIPE')
            if stderr is not None:
                if stderr == DEVNULL:
                    command.append('stderr=0')
                elif stderr == STDOUT:
                    command.append('stderr=stdout')
                elif stderr != PIPE:
                    raise utils.Error('stderr must be DEVNULL, PIPE or STDOUT')
        if env:
            for key, value in env.items():
                command.append('env.{0}={1}'.format(key, wrap_trailing_slash(value)))
        if rlimits:
            for key, value in rlimits.items():
                command.append('rlimit.{0}={1}'.format(key.lower(), value))
        if executable:
            command.append(executable)
        if args:
            if isinstance(args, string_type):
                command.append(args)
            else:
                try:
                    command.extend(args)
                except TypeError:
                    command.append(str(args))
        if watch_immediately:
            return self._do_watch(command)
        return self._make_process_dict(self._do_command(command))[1]

    def remove_processes(self, *args):
        self._do_command(['r', 'p'] + list(args))
        return True

    def watch_processes(self, *args):
        return self._do_watch(['w', 'p'] + list(args))

    def exit_if_idle(self):
        return self._do_command('exit-if-idle').startswith('Exiting')

    def _do_watch(self, command):
        self._send_command(command)
        self.connection.get_one_line_response()
        watcher = Watcher(self)
        self.connection = None
        return watcher

    def close(self):
        if self.connection:
            self.connection.close()
            self.connection = None

    @classmethod
    def set_debug_mode(cls, mode=True, quit_when_connection_closed=False):
        cls._debug_mode = mode
        if quit_when_connection_closed:
            cls._single_connection_mode = True

    @classmethod
    def set_default_port(cls, port):
        cls._default_port_or_path = port

    @classmethod
    def set_default_path(cls, path):
        cls._default_port_or_path = path


def set_debug_mode(mode=True, quit_when_connection_closed=False):
    return Papa.set_debug_mode(mode, quit_when_connection_closed)


def set_default_port(port):
    return Papa.set_default_port(port)


def set_default_path(path):
    return Papa.set_default_path(path)

from papa import utils
s = utils.cast_string
b = utils.cast_bytes
Error = utils.Error

if __name__ == '__main__':
    set_debug_mode(quit_when_connection_closed=True)
    p = Papa()
    # print('Sockets: {0}'.format(p.list_sockets()))
    # print('Socket uwsgi: {0}'.format(p.make_socket('uwsgi', interface='eth0')))
    # print('Sockets: {0}'.format(p.list_sockets()))
    # print('Socket chaussette: {0}'.format(p.make_socket('chaussette', path='/tmp/chaussette.sock')))
    # print('Sockets: {0}'.format(p.list_sockets()))
    # print('Socket uwsgi6: {0}'.format(p.make_socket('uwsgi6', family=socket.AF_INET6)))
    # print('Sockets: {0}'.format(p.list_sockets()))
    # try:
    #     print('Socket chaussette: {0}'.format(p.make_socket('chaussette')))
    # except Exception as e:
    #     print('Caught exception: {0}'.format(e))
    # print('Close chaussette: {0}'.format(p.remove_sockets('chaussette')))
    # print('Socket chaussette: {0}'.format(p.make_socket('chaussette')))
    # print('Sockets: {0}'.format(p.list_sockets()))
    print('Processes: {0}'.format(p.list_processes()))
    # print('Values: {0}'.format(p.list_values()))
    # print('Set aack: {0}'.format(p.set('aack', 'bar')))
    # print('Get aack: {0}'.format(p.get('aack')))
    # print('Get bar: {0}'.format(p.get('bar')))
    # print('Values: {0}'.format(p.list_values()))
    # print('Set bar: {0}'.format(p.set('bar', 'barry')))
    # print('Get bar: {0}'.format(p.get('bar')))
    # print('Values: {0}'.format(p.list_values()))
    # print('Set bar: {0}'.format(p.set('bar')))
    # print('Get bar: {0}'.format(p.get('bar')))
    # print('Values: {0}'.format(p.list_values()))
    # print('Set aack: {0}'.format(p.set('aack')))
    p.close()
    print('Killed papa')
